@@ -8,6 +8,9 @@
    ##################################################################### */
 									/*}}}*/
 // Include Files							/*{{{*/
+#include <config.h>
+
+#include <apt-pkg/aptconfiguration.h>
 #include <apt-pkg/fileutl.h>
 #include <apt-pkg/acquire-method.h>
 #include <apt-pkg/acquire-item.h>
@@ -15,19 +18,26 @@
 #include <apt-pkg/error.h>
 #include <apt-pkg/hashes.h>
 #include <apt-pkg/sourcelist.h>
+#include <apt-pkg/configuration.h>
+#include <apt-pkg/metaindex.h>
 
+#include <algorithm>
 #include <fstream>
 #include <iostream>
+
 #include <stdarg.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/utsname.h>
 #include <dirent.h>
 
 using namespace std;
 
+#include<sstream>
+
 #include "mirror.h"
 #include "http.h"
-#include "apti18n.h"
+#include <apti18n.h>
 									/*}}}*/
 
 /* Done:
@@ -48,7 +58,7 @@ using namespace std;
  */
 
 MirrorMethod::MirrorMethod()
-   : HttpMethod(), DownloadedMirrorFile(false)
+   : HttpMethod(), DownloadedMirrorFile(false), Debug(false)
 {
 };
 
@@ -101,10 +111,10 @@ bool MirrorMethod::Clean(string Dir)
 	 continue;
 
       // see if we have that uri
-      for(I=list.begin(); I != list.end(); I++)
+      for(I=list.begin(); I != list.end(); ++I)
       {
 	 string uri = (*I)->GetURI();
-	 if(uri.substr(0,strlen("mirror://")) != string("mirror://"))
+	 if(uri.find("mirror://") != 0)
 	    continue;
 	 string BaseUri = uri.substr(0,uri.size()-1);
 	 if (URItoFileName(BaseUri) == Dir->d_name)
@@ -114,53 +124,153 @@ bool MirrorMethod::Clean(string Dir)
       if (I == list.end())
 	 unlink(Dir->d_name);
    };
-   
-   chdir(StartDir.c_str());
+
    closedir(D);
+   if (chdir(StartDir.c_str()) != 0)
+      return _error->Errno("chdir",_("Unable to change to %s"),StartDir.c_str());
    return true;   
 }
 
 
 bool MirrorMethod::DownloadMirrorFile(string mirror_uri_str)
 {
-
-   // check the file, if it is not older than RefreshInterval just use it
-   // otherwise try to get a new one
-   if(FileExists(MirrorFile)) 
-   {
-      struct stat buf;
-      time_t t,now,refresh;
-      if(stat(MirrorFile.c_str(), &buf) != 0)
-	 return false;
-      t = std::max(buf.st_mtime, buf.st_ctime);
-      now = time(NULL);
-      refresh = 60*_config->FindI("Acquire::Mirror::RefreshInterval",360);
-      if(t + refresh > now)
-      {
-	 if(Debug)
-	    clog << "Mirror file is in RefreshInterval" << endl;
-	 DownloadedMirrorFile = true;
-	 return true;
-      }
-      if(Debug)
-	 clog << "Mirror file " << MirrorFile << " older than " << refresh << "min, re-download it" << endl;
-   }
-
    // not that great to use pkgAcquire here, but we do not have 
    // any other way right now
    string fetch = BaseUri;
    fetch.replace(0,strlen("mirror://"),"http://");
 
+#if 0 // no need for this, the getArchitectures() will also include the main 
+      // arch
+   // append main architecture
+   fetch += "?arch=" + _config->Find("Apt::Architecture");
+#endif
+
+   // append all architectures
+   std::vector<std::string> vec = APT::Configuration::getArchitectures();
+   for (std::vector<std::string>::const_iterator I = vec.begin();
+        I != vec.end(); ++I)
+      if (I == vec.begin())
+         fetch += "?arch=" + (*I);
+      else
+         fetch += "&arch=" + (*I);
+
+   // append the dist as a query string
+   if (Dist != "")
+      fetch += "&dist=" + Dist;
+
+   if(Debug)
+      clog << "MirrorMethod::DownloadMirrorFile(): '" << fetch << "'"
+           << " to " << MirrorFile << endl;
+
    pkgAcquire Fetcher;
    new pkgAcqFile(&Fetcher, fetch, "", 0, "", "", "", MirrorFile);
    bool res = (Fetcher.Run() == pkgAcquire::Continue);
-   if(res)
+   if(res) {
       DownloadedMirrorFile = true;
+      chmod(MirrorFile.c_str(), 0644);
+   }
    Fetcher.Shutdown();
+
+   if(Debug)
+      clog << "MirrorMethod::DownloadMirrorFile() success: " << res << endl;
+   
    return res;
 }
 
-bool MirrorMethod::SelectMirror()
+// Randomizes the lines in the mirror file, this is used so that
+// we spread the load on the mirrors evenly
+bool MirrorMethod::RandomizeMirrorFile(string mirror_file)
+{
+   vector<string> content;
+   string line;
+
+   if (!FileExists(mirror_file))
+      return false;
+
+   // read 
+   ifstream in(mirror_file.c_str());
+   while ( !in.eof() ) {
+      getline(in, line);
+      content.push_back(line);
+   }
+   
+   // we want the file to be random for each different machine, but also
+   // "stable" on the same machine. this is to avoid running into out-of-sync
+   // issues (i.e. Release/Release.gpg different on each mirror)
+   struct utsname buf;
+   int seed=1, i;
+   if(uname(&buf) == 0) {
+      for(i=0,seed=1; buf.nodename[i] != 0; i++) {
+         seed = seed * 31 + buf.nodename[i];
+      }
+   }
+   srand( seed );
+   random_shuffle(content.begin(), content.end());
+
+   // write
+   ofstream out(mirror_file.c_str());
+   while ( !content.empty()) {
+      line = content.back();
+      content.pop_back();
+      out << line << "\n";
+   }
+
+   return true;
+}
+
+/* convert a the Queue->Uri back to the mirror base uri and look
+ * at all mirrors we have for this, this is needed as queue->uri
+ * may point to different mirrors (if TryNextMirror() was run)
+ */
+void MirrorMethod::CurrentQueueUriToMirror()
+{
+   // already in mirror:// style so nothing to do
+   if(Queue->Uri.find("mirror://") == 0)
+      return;
+
+   // find current mirror and select next one
+   for (vector<string>::const_iterator mirror = AllMirrors.begin();
+	mirror != AllMirrors.end(); ++mirror)
+   {
+      if (Queue->Uri.find(*mirror) == 0)
+      {
+	 Queue->Uri.replace(0, mirror->length(), BaseUri);
+	 return;
+      }
+   }
+   _error->Error("Internal error: Failed to convert %s back to %s",
+		 Queue->Uri.c_str(), BaseUri.c_str());
+}
+
+bool MirrorMethod::TryNextMirror()
+{
+   // find current mirror and select next one
+   for (vector<string>::const_iterator mirror = AllMirrors.begin();
+	mirror != AllMirrors.end(); ++mirror)
+   {
+      if (Queue->Uri.find(*mirror) != 0)
+	 continue;
+
+      vector<string>::const_iterator nextmirror = mirror + 1;
+      if (nextmirror == AllMirrors.end())
+	 break;
+      Queue->Uri.replace(0, mirror->length(), *nextmirror);
+      if (Debug)
+	 clog << "TryNextMirror: " << Queue->Uri << endl;
+
+      // inform parent
+      UsedMirror = *nextmirror;
+      Log("Switching mirror");
+      return true;
+   }
+
+   if (Debug)
+      clog << "TryNextMirror could not find another mirror to try" << endl;
+
+   return false;
+}
+
+bool MirrorMethod::InitMirrors()
 {
    // if we do not have a MirrorFile, fallback
    if(!FileExists(MirrorFile))
@@ -170,6 +280,13 @@ bool MirrorMethod::SelectMirror()
       return _error->Error(_("No mirror file '%s' found "), MirrorFile.c_str());
    }
 
+   if (access(MirrorFile.c_str(), R_OK) != 0)
+   {
+      // FIXME: fallback to a default mirror here instead 
+      //        and provide a config option to define that default
+      return _error->Error(_("Can not read mirror file '%s'"), MirrorFile.c_str());
+   }  
+
    // FIXME: make the mirror selection more clever, do not 
    //        just use the first one!
    // BUT: we can not make this random, the mirror has to be
@@ -177,10 +294,24 @@ bool MirrorMethod::SelectMirror()
    //      get into sync issues (got indexfiles from mirror A,
    //      but packages from mirror B - one might be out of date etc)
    ifstream in(MirrorFile.c_str());
-   getline(in, Mirror);
-   if(Debug)
-      cerr << "Using mirror: " << Mirror << endl;
+   string s;
+   while (!in.eof()) 
+   {
+      getline(in, s);
 
+      // ignore lines that start with #
+      if (s.find("#") == 0)
+         continue;
+      // ignore empty lines
+      if (s.size() == 0)
+         continue;
+      // ignore non http lines
+      if (s.find("http://") != 0)
+         continue;
+
+      AllMirrors.push_back(s);
+   }
+   Mirror = AllMirrors[0];
    UsedMirror = Mirror;
    return true;
 }
@@ -219,7 +350,7 @@ string MirrorMethod::GetMirrorFileName(string mirror_uri_str)
    vector<metaIndex *>::const_iterator I;
    pkgSourceList list;
    list.ReadMainList();
-   for(I=list.begin(); I != list.end(); I++)
+   for(I=list.begin(); I != list.end(); ++I)
    {
       string uristr = (*I)->GetURI();
       if(Debug)
@@ -232,6 +363,7 @@ string MirrorMethod::GetMirrorFileName(string mirror_uri_str)
 	 if(Debug)
 	    std::cerr << "found BaseURI: " << uristr << std::endl;
 	 BaseUri = uristr.substr(0,uristr.size()-1);
+         Dist = (*I)->GetDist();
       }
    }
    // get new file
@@ -251,6 +383,9 @@ string MirrorMethod::GetMirrorFileName(string mirror_uri_str)
    depth. */
 bool MirrorMethod::Fetch(FetchItem *Itm)
 {
+   if(Debug)
+      clog << "MirrorMethod::Fetch()" << endl;
+
    // the http method uses Fetch(0) as a way to update the pipeline,
    // just let it do its work in this case - Fetch() with a valid
    // Itm will always run before the first Fetch(0)
@@ -266,17 +401,24 @@ bool MirrorMethod::Fetch(FetchItem *Itm)
    if(Itm->IndexFile && !DownloadedMirrorFile)
    {
       Clean(_config->FindDir("Dir::State::mirrors"));
-      DownloadMirrorFile(Itm->Uri);
+      if (DownloadMirrorFile(Itm->Uri))
+         RandomizeMirrorFile(MirrorFile);
    }
 
-   if(Mirror.empty())
-      SelectMirror();
-
-   for (FetchItem *I = Queue; I != 0; I = I->Next)
-   {
-      if(I->Uri.find("mirror://") != string::npos)
-	 I->Uri.replace(0,BaseUri.size(), Mirror);
+   if(AllMirrors.empty()) {
+      if(!InitMirrors()) {
+	 // no valid mirror selected, something went wrong downloading
+	 // from the master mirror site most likely and there is
+	 // no old mirror file availalbe
+	 return false;
+      }
    }
+
+   if(Itm->Uri.find("mirror://") != string::npos)
+      Itm->Uri.replace(0,BaseUri.size(), Mirror);
+
+   if(Debug)
+      clog << "Fetch: " << Itm->Uri << endl << endl;
    
    // now run the real fetcher
    return HttpMethod::Fetch(Itm);
@@ -284,22 +426,35 @@ bool MirrorMethod::Fetch(FetchItem *Itm)
 
 void MirrorMethod::Fail(string Err,bool Transient)
 {
-   if(Queue->Uri.find("http://") != string::npos)
-      Queue->Uri.replace(0,Mirror.size(), BaseUri);
+   // FIXME: TryNextMirror is not ideal for indexfile as we may
+   //        run into auth issues
+
+   if (Debug)
+      clog << "Failure to get " << Queue->Uri << endl;
+
+   // try the next mirror on fail (if its not a expected failure,
+   // e.g. translations are ok to ignore)
+   if (!Queue->FailIgnore && TryNextMirror()) 
+      return;
+
+   // all mirrors failed, so bail out
+   string s;
+   strprintf(s, _("[Mirror: %s]"), Mirror.c_str());
+   SetIP(s);
+
+   CurrentQueueUriToMirror();
    pkgAcqMethod::Fail(Err, Transient);
 }
 
 void MirrorMethod::URIStart(FetchResult &Res)
 {
-   if(Queue->Uri.find("http://") != string::npos)
-      Queue->Uri.replace(0,Mirror.size(), BaseUri);
+   CurrentQueueUriToMirror();
    pkgAcqMethod::URIStart(Res);
 }
 
 void MirrorMethod::URIDone(FetchResult &Res,FetchResult *Alt)
 {
-   if(Queue->Uri.find("http://") != string::npos)
-      Queue->Uri.replace(0,Mirror.size(), BaseUri);
+   CurrentQueueUriToMirror();
    pkgAcqMethod::URIDone(Res, Alt);
 }
 
